#include <u.h>
#include <libc.h>
#include <aml.h>

/* taken from aml.c, should be exported by aml.h, but well... */
enum {
	MemSpace	= 0x00,
	IoSpace		= 0x01,
	PcicfgSpace	= 0x02,
	EbctlSpace	= 0x03,
	SmbusSpace	= 0x04,
	CmosSpace	= 0x05,
	PcibarSpace	= 0x06,
	IpmiSpace	= 0x07,
};

struct ec {
	int fd;
	uchar (*read)(uchar);
	void (*write)(uchar, uvlong);
} acpiec;

typedef struct Tbl Tbl;

struct Tbl {
	uchar	sig[4];
	uchar	len[4];
	uchar	rev;
	uchar	csum;
	uchar	oemid[6];
	uchar	oemtid[8];
	uchar	oemrev[4];
	uchar	cid[4];
	uchar	crev[4];
	uchar	data[];
};

static Tbl *tbltab[64];

uchar
ecread(uchar addr) {
	uvlong v;

	if(pread(acpiec.fd, &v, 1, addr) == -1)
		exits("ecread: can't read");
	return v;
}

void
ecwrite(uchar addr, uvlong val) {
	if(pwrite(acpiec.fd, &val, sizeof(uvlong), addr) == -1)
		exits("ecwrite: can't write");
}

void
ecdummywrite(uchar, uvlong) {
}

void
ecinit(void) {
	int fd;

	if((fd = open("/dev/acpiec", ORDWR)) != -1) {
		acpiec.fd = fd;
		acpiec.read = ecread;
		acpiec.write = ecwrite;
	} else {
		exits("acpiec not available");
	}
}

void 
acpi_read(int space, int off, int, uvlong *buf)
{
	int n;

	if (space != EbctlSpace) {
		*buf = 0;
		return;
	}
	//print("acpiec_read: off 0x%x len %d\n", off, len);

	n = acpiec.read(off);
	*buf = n;
}

void 
acpi_write(int space, int off, int, uvlong val)
{
	if (space != EbctlSpace) {
		return;
	}
	//print("acpiec_write: off 0x%x len %d\n", off, len);
	acpiec.write(off, val);
}

void 
acpi_delay(int n) 
{
	sleep(n);
}

void*
amlalloc(int n){
	void *p;

	if((p = malloc(n)) == nil) 
		sysfatal("amlalloc: no memory");
	memset(p, 0, n);
	return p;
}

void
amlfree(void *p){
	free(p);
}

void*
gettables(int *len) {
	int fd, n, size;
	uvlong off;
	void *tables;
	uchar *p;

	size = 1024*1024;
	tables = malloc(size);
	p = (uchar*)tables;
	if((fd = open("/dev/acpitbls", OREAD)) != -1) {
		off = 0;
		while((n = read(fd, p, 1024)) > 0) {
			off += n;
			p += n;
			if(off >= size - 1024) {
				size *= 2;
				tables = realloc(tables, size);
				p = (uchar*)tables + off;
			}
		}
	} else {
		exits("acpi tables not available");
	}
	*len = off;
	close(fd);
	return tables;
}

Tbl*
findtable(uchar *tp, char name[4]) {
	int i;
	Tbl *t;

	for(i = 0; i < nelem(tbltab); i++) {
		t = (Tbl*)tp;
		if(memcmp(t->sig, name, 4) == 0)
			return t;
		tp += *((uint*)t->len);
	}
	return nil;
}

static uint
tbldlen(Tbl *t){
	return *((uint*)t->len) - sizeof(Tbl);
}

static int
enumini(void *dot, void *)
{
	void *m;
	void **p;
	uvlong sta;

	enum { 
		StaPresent = 1 << 0, 
		StaEnabled = 1 << 1,
		StaDevOk = 1 << 3 
	};

	if((m = amlwalk(dot, "^_STA")) == 0) {
		/* no _STA, call _INI */
		amleval(dot, "", nil);
		return 0;
	} else if(amleval(m, "", &p) < 0) {
		print("ini: _STA not working\n");
		return -1;
	}
	sta = amlint(p);
	if (sta & StaPresent) {
		/* _STA says present, call _INI */
		amleval(dot, "", nil);
	}
		
	return 0;
}

static char*
eisaid(void *v)
{
	static char id[8];
	ulong b, l;
	int i;

	if(amltag(v) == 's')
		return v;
	b = amlint(v);
	for(l = 0, i=24; i>=0; i -= 8, b >>= 8)
		l |= (b & 0xFF) << i;
	id[7] = 0;
	for(i=6; i>=3; i--, l >>= 4)
		id[i] = "0123456789ABCDEF"[l & 0xF];
	for(i=2; i>=0; i--, l >>= 5)
		id[i] = '@' + (l & 0x1F);
	return id;
}

static int
acpiacstatus(void *mi) 
{
	void **p;
	static void *m = nil;

	if(mi != nil)
		m = mi;

	if(m == nil) {
		print("must not happen\n");
		return 0;
	}
	if(amleval(m, "", &p) < 0) {
		print("ac: _PSR not working\n");
		return 0;
	}
	return amlint(p);
}

static void 
acpibat(void *dot) 
{       
        uvlong sta; 
        void *m; 
        void *p, **a; 
        enum { BatteryPresent = (1L << 4) }; 
        enum { BifUnknown = 0xffffffff }; 
 
        if((m = amlwalk(dot, "_STA")) == 0) { 
                 print("bat: no _STA\n"); 
                 return; 
        } 
        if(amleval(m, "", &p) < 0) { 
                 print("bat: _STA not working\n"); 
                 return; 
        } 
        sta = amlint(p); 
        if((sta & BatteryPresent) != 0) { 
                 print("bat:\n"); 
 
                 if((m = amlwalk(dot, "_BIF")) == 0) { 
                          print("bat: no _BIF\n"); 
                          return; 
                 } 
                 if(amleval(m, "", &p) < 0) { 
                          print("bat: _BIF not working\n"); 
                          return; 
                 } 
                 a = amlval(p); 
                 print("\tpower unit: %ulld\n", amlint(a[0])); 
                 print("\tcapacity: %ulld\n", amlint(a[1])); 
                 print("\tlast capacity: %ulld\n", amlint(a[2])); 
                 print("\tvoltage: %ulld\n", amlint(a[4])); 
                 print("\tmodel: %s\n", (char*)amlval(a[9])); 
                 print("\ttype: %s\n", (char*)amlval(a[11])); 
                 print("\toem: %s\n", (char*)amlval(a[12])); 
 
                 if((m = amlwalk(dot, "_BST")) == 0) { 
                          print("bat: no _BST\n"); 
                          return; 
                 } 
                 if(amleval(m, "", &p) < 0) { 
                          print("bat: _BST not working\n"); 
                          return; 
                 }        
                 a = amlval(p); 
                 print("\tcur.state: %ulld\n", amlint(a[0])); 
                 print("\tcur.rate: %ulld\n", amlint(a[1])); 
                 print("\tcur.capacity: %ulld\n", amlint(a[2])); 
                 print("\tcur.voltage: %ulld\n", amlint(a[3])); 
        } 
} 

static void
acpiac(void *dot) 
{
	uvlong sta;
	void *m;
	void *p;
	enum { Offline, Online };

	if((m = amlwalk(dot, "_STA")) == 0) {
		print("ac: no _STA\n");
		//return;
	} else if(amleval(m, "", &p) < 0) {
		print("ac: _STA not working\n");
		return;
	}
	sta = amlint(p);
	if((m = amlwalk(dot, "_PSR")) == 0) {
		print("ac: no _PSR\n");
		return;
	}
	sta = acpiacstatus(m);
	if(sta == Online)
		print("ac: AC online\n");
	else if(sta == Offline)
		print("ac: AC offline\n");
	else
		print("ac: status unknown\n");
}

static int 
enumhid(void *dot, void *)
{
	void *p;
	char *id;

	id = nil;
	p = nil;
	if(amleval(dot, "", &p) == 0)
		id = eisaid(p);

	if(id && (!strcmp(id, "PNP0C0A")))
		acpibat(dot);

	if(id && (!strcmp(id, "ACPI0003")))
		acpiac(dot);
	return -1;
}

void
main() {
	uchar *p;
	int len;
	uvlong val;
	Tbl *t;

	ecinit();
	p = gettables(&len);
	print("tables @ 0x%x, len = %d\n", p, len);

	amlinit();

	if(t = findtable(p, "DSDT")) {
		amlload(t->data, tbldlen(t));
	}
	if(t = findtable(p, "SSDT")) {
		amlload(t->data, tbldlen(t));
	}

	/* 
	 * Bad things happen when doing actual writes
	 */
	acpiec.write = ecdummywrite;
	amlenum(amlroot, "_INI", enumini, nil);
	acpiec.write = ecwrite;

	/* look for other ACPI devices */
	amlenum(amlroot, "_HID", enumhid, nil);

	exits("");
} 